/*=========================================================================*/
/*  for MultiCPU(v2)                                                       */
/*=========================================================================*/
struct CompressUnit {
	CodecInst* pCodec;
	DWORD retval;
	ICCOMPRESS* icinfo;
	DWORD dwSize;
	DWORD height_st;
	DWORD height;
	DWORD offset;
	DWORD size;
};

struct DecompressUnit {
	CodecInst* pCodec;
	DWORD retval;
	ICDECOMPRESS* icinfo;
	DWORD dwSize;
	DWORD height_st;
	DWORD height;
	DWORD offset;
	DWORD size;
};

//static int GetBitmapType(LPBITMAPINFOHEADER lpbi)
// 0=unknown, -1=compressed YUY2, -2=compressed RGB, -3=compressed RGBA, 1=YUY2, 2=UYVY, 3=RGB 24-bit, 4=RGB 32-bit
static int GetByteCount(LPBITMAPINFOHEADER lpbi) {
  if (!lpbi)
    return 0;
  const int fourcc = lpbi->biCompression;
  if (fourcc == FOURCC_VYUY || fourcc == FOURCC_YUY2)
    return 2;
  if (fourcc == FOURCC_UYVY)
    return 2;
  const int bitcount = GetBitCount(lpbi);
  if (fourcc == 0 || fourcc == ' BID')
    return (bitcount == 24) ? 3 : (bitcount == 32) ? 4 : 0;
  if (fourcc == FOURCC_HFYU)
    return ((bitcount&~7) == 16) ? 2 : ((bitcount&~7) == 24) ? 3 : ((bitcount&~7) == 32) ? 4 : 0;
  return 0;
}


static DWORD ConvertYUY2toRGB2(LPBITMAPINFOHEADER lpbiOutput, void* _src, void* _dst, unsigned src_height) {
  const unsigned char* src = (const unsigned char*)_src;
  unsigned char* dst = (unsigned char*)_dst;
  int stride = lpbiOutput->biWidth * 2;
  //const unsigned char* src_end = src + stride*lpbiOutput->biHeight;
  const unsigned char* src_end = src + stride*src_height;
  if (lpbiOutput->biBitCount == 32)
    mmx_YUY2toRGB32(src, dst, src_end, stride);
  else
    mmx_YUY2toRGB24(src, dst, src_end, stride);
  return ICERR_OK;
}

static DWORD ConvertUYVYtoRGB2(LPBITMAPINFOHEADER lpbiOutput, void* _src, void* _dst, unsigned src_height) {
  const unsigned char* src = (const unsigned char*)_src;
  unsigned char* dst = (unsigned char*)_dst;
  int stride = lpbiOutput->biWidth * 2;
  //const unsigned char* src_end = src + stride*lpbiOutput->biHeight;
  const unsigned char* src_end = src + stride*src_height;
  if (lpbiOutput->biBitCount == 32)
    mmx_UYVYtoRGB32(src, dst, src_end, stride);
  else
    mmx_UYVYtoRGB24(src, dst, src_end, stride);
  return ICERR_OK;
}

/*=========================================================================*/
/*  Compress for MultiCPU(v2)                                              */
/*=========================================================================*/
unsigned int __stdcall CodecInst::Compress3_thd(void* args)
{
	DBGPRINTOUT("---Start Compress3Thread---\n")
	CompressUnit *thd = (CompressUnit *)args;
	thd->retval = (thd->pCodec)->Compress3_Core2(thd->icinfo, thd->dwSize, thd->height_st, thd->height, &(thd->offset), &(thd->size));
	DBGPRINTOUT("---End Compress3Thread---\n")
	_endthreadex(0);
	return 0;
}

DWORD CodecInst::Compress3_Core1(ICCOMPRESS* icinfo, DWORD dwSize) {
  
  if (icinfo->lpckid)
    *icinfo->lpckid = FOURCC_HFYU;
  *icinfo->lpdwFlags = AVIIF_KEYFRAME;

  return ICERR_OK;
}

DWORD CodecInst::Compress3_Core2(ICCOMPRESS* icinfo, DWORD dwSize, DWORD in_height_st, DWORD in_height, LPDWORD out_st, LPDWORD out_size) {
  //------------------------------------------------------------------------------
  //Note: [out_st] in:out_height_st, out:out_byte_offset
  LONG in_biHeight = in_height;
  DWORD in_offs  = icinfo->lpbiInput->biWidth * in_height_st * GetByteCount(icinfo->lpbiInput);
  DWORD out_offs = icinfo->lpbiInput->biWidth * (*out_st) * GetByteCount(icinfo->lpbiOutput);
  unsigned char* u_yuy2_buffer   = yuy2_buffer+out_offs;
  unsigned char* u_median_buffer = median_buffer+out_offs;
  unsigned char* u_rgb_buffer    = rgb_buffer+out_offs;
  //------------------------------------------------------------------------------

  int intype = GetBitmapType(icinfo->lpbiInput);
  int outtype = GetBitmapType(icinfo->lpbiOutput);

  if (outtype < 0) {    // compressing (as opposed to converting RGB->YUY2)
    const int method = GetMethod(icinfo->lpbiOutput);

    if (encode_table_owner != this) {
      InitializeEncodeTables(GetEmbeddedHuffmanTable(icinfo->lpbiOutput));
      encode_table_owner = this;
    }

    const unsigned char* const in = (unsigned char*)icinfo->lpInput + in_offs;
    unsigned long* const out = (unsigned long*)icinfo->lpOutput + (out_offs/sizeof(unsigned long));

    unsigned long* out_end;

    if (outtype == -1) {    // compressing to HFYU16 (4:2:2 YUV)
      int stride = icinfo->lpbiInput->biWidth * 2;
      int size = stride * in_biHeight;
      if (icinfo->lpbiInput->biHeight > field_threshold) stride *= 2;    // if image is interlaced, double stride so fields are treated separately
      const unsigned char* yuy2in = in;

      if (intype == 3) {    // RGB24->YUY2->HFYU16
        ConvertRGB24toYUY2(in, u_yuy2_buffer, icinfo->lpbiInput->biWidth, in_biHeight);
        yuy2in = u_yuy2_buffer;
        size = icinfo->lpbiInput->biWidth * in_biHeight * 2;
      }
      if (method == methodGrad) {
        // Attempt to match yuy2_buffer alignment to input alignment.
        // Why, oh why, can't standard libraries just quadword-align
        // big memory blocks and be done with it?
        unsigned char* const aligned_yuy2_buffer = u_yuy2_buffer + 4*((int(yuy2in)&7) != (int(u_yuy2_buffer)&7));
        mmx_RowDiff(yuy2in, aligned_yuy2_buffer, yuy2in+size, stride);
        yuy2in = aligned_yuy2_buffer;
      }
      if (intype==2) {    // UYVY->HFYU16
        if (method == methodMedian) {
          unsigned char* aligned_median_buffer = Align8(u_median_buffer);
          mmx_MedianPredictUYVY(yuy2in, aligned_median_buffer, yuy2in+size, stride);
          out_end = asm_CompressUYVY(aligned_median_buffer, out, aligned_median_buffer+size);
        }
        else {
          out_end = asm_CompressUYVYDelta(yuy2in,out,yuy2in+size);
        }
      } else {    // YUY2->HFYU16
        if (method == methodMedian) {
          unsigned char* aligned_median_buffer = Align8(u_median_buffer);
          mmx_MedianPredictYUY2(yuy2in, aligned_median_buffer, yuy2in+size, stride);
          out_end = asm_CompressYUY2(aligned_median_buffer, out, aligned_median_buffer+size);
        }
        else {
          out_end = asm_CompressYUY2Delta(yuy2in,out,yuy2in+size);
        }
      }
    }
    else {    // compressing to HFYU24 (RGB) or HFYU32 (RGBA)
      int stride = (icinfo->lpbiInput->biWidth * GetBitCount(icinfo->lpbiInput)) >> 3;
      int size = stride * in_biHeight;
      if (icinfo->lpbiInput->biHeight > field_threshold) stride *= 2;    // if image is interlaced, double stride so fields are treated separately
      const unsigned char* rgbin = in;

      if ((method&~flagDecorrelate) == methodGrad) {
        unsigned char* const aligned_rgb_buffer = u_rgb_buffer + 4*((int(in)&7) != (int(u_rgb_buffer)&7));
        mmx_RowDiff(in, aligned_rgb_buffer, in+size, stride);
        rgbin = aligned_rgb_buffer;
      }
      if (outtype == -2) {    // RGB24->HFYU24
        if (method == methodLeft || method == methodOld)
          out_end = asm_CompressRGBDelta(rgbin,out,rgbin+size);
        else
          out_end = asm_CompressRGBDeltaDecorrelate(rgbin,out,rgbin+size);
      }
      else if (outtype == -3) {     // RGBA->HFYU32
        if (method == methodLeft || method == methodOld)
          out_end = asm_CompressRGBADelta(rgbin,out,rgbin+size);
        else
          out_end = asm_CompressRGBADeltaDecorrelate(rgbin,out,rgbin+size);
      }
      else
        return ICERR_BADFORMAT;
    }
	*out_st   = out_offs;
	*out_size = DWORD(out_end) - DWORD(out);
    //icinfo->lpbiOutput->biSizeImage = DWORD(out_end) - DWORD(out);
    return ICERR_OK;
  }
  else if (outtype == 1) {    // RGB24->YUY2
    ConvertRGB24toYUY2((unsigned char*)icinfo->lpInput+in_offs, (unsigned char*)icinfo->lpOutput+out_offs, icinfo->lpbiInput->biWidth, in_biHeight);
	*out_st   = out_offs;
	*out_size = (icinfo->lpbiOutput->biWidth * in_biHeight * GetBitCount(icinfo->lpbiOutput)) >> 3;
    //icinfo->lpbiOutput->biSizeImage = (icinfo->lpbiOutput->biWidth * icinfo->lpbiOutput->biHeight * GetBitCount(icinfo->lpbiOutput)) >> 3;
    return ICERR_OK;
  }
  else
    return ICERR_BADFORMAT;
}


DWORD CodecInst::Compress3(ICCOMPRESS* icinfo, DWORD dwSize) {
	DWORD retval = ICERR_OK;

	int cut_num = (forced_processors) ? forced_processors : processors;
	if(cut_num > MAX_PROCESSOR_NUM) return ICERR_BADFORMAT;
	//Msg("---Compress3[num=%d]---\n", cut_num);

	//---Ready compress
	retval = Compress3_Core1(icinfo, dwSize);
    if (retval != ICERR_OK)
		return retval;

	//---Start threads
	int i;
	int numThreads = 0;
	HANDLE hThread[MAX_PROCESSOR_NUM];
	DWORD height_num = icinfo->lpbiInput->biHeight / cut_num;
	height_num = height_num + (height_num%2==0 ? 0 : 1);
	CompressUnit comunit[MAX_PROCESSOR_NUM];
	bool rgb = false;
	if( (GetBitmapType(icinfo->lpbiOutput)==1) || (GetBitmapType(icinfo->lpbiOutput)==-1 && GetBitmapType(icinfo->lpbiInput)==3) )
		rgb = true;
	for(i=0; i<cut_num; i++){
		comunit[i].pCodec    = this;
		comunit[i].retval    = ICERR_OK;
		comunit[i].icinfo    = icinfo;
		comunit[i].dwSize    = dwSize;
		
		if(!rgb){
			comunit[i].height_st = height_num*i;
		} else {
			comunit[i].height_st = (icinfo->lpbiInput->biHeight) - (height_num * (i+1));
		}
		comunit[i].height    = height_num;
		comunit[i].offset    = height_num*i;
		comunit[i].size      = 0;
		if(i==cut_num-1){
			comunit[i].height = icinfo->lpbiInput->biHeight - (height_num*i);
			if(rgb) comunit[i].height_st = 0;
		}
		hThread[numThreads] = (HANDLE)_beginthreadex(NULL, 0, Compress3_thd, &comunit[i], 0, NULL);
		SetThreadAffinityMask(hThread[numThreads], processor_mask[numThreads]);
		numThreads++;
		//retval = Compress3_Core2(icinfo, dwSize, comunit[i].height_st, comunit[i].height, &comunit[i].offset, &comunit[i].size);
		//if (retval != ICERR_OK)
		//	return retval;
	}
	//---Wait multiple thread finish
	WaitForMultipleObjects(numThreads, hThread, TRUE, INFINITE);
	for(i=0; i<numThreads; i++){
		CloseHandle(hThread[i]);
		if(comunit[i].retval != ICERR_OK)	
			retval = comunit[i].retval;
	}
	if(retval != ICERR_OK)	return retval;

	//---Concatenate data
	icinfo->lpbiOutput->biSizeImage = comunit[0].size;
	for(i=1; i<cut_num; i++){
		unsigned long srcoffs = comunit[i].offset;
		unsigned char* src    = (unsigned char*)icinfo->lpOutput + srcoffs;
		unsigned long srcsize = comunit[i].size;
		unsigned long dstoffs = comunit[i-1].offset + comunit[i-1].size;
		unsigned char* dst    = (unsigned char*)icinfo->lpOutput + dstoffs;
		memmove(dst, src, srcsize);
		comunit[i].offset = dstoffs;
		icinfo->lpbiOutput->biSizeImage += srcsize;
	}
	//---Add Infomation data
	unsigned long codesize = icinfo->lpbiOutput->biSizeImage;
	CompressInfo* info = (CompressInfo*)((unsigned char*)(icinfo->lpOutput) + icinfo->lpbiOutput->biSizeImage);
	for(i=0; i<cut_num; i++){
		info[i].code_start = comunit[i].offset;		//[start offset]
		info[i].data_size  = comunit[i].size;		//[data size]
		icinfo->lpbiOutput->biSizeImage += sizeof(CompressInfo);
	}
	CompressInfoHeader* header = (CompressInfoHeader*)((unsigned char*)(icinfo->lpOutput) + icinfo->lpbiOutput->biSizeImage);
	header->id       = 0x00;
	header->unitnum  = cut_num;
	header->unitsize = height_num;
	header->offset   = codesize;
	icinfo->lpbiOutput->biSizeImage += sizeof(CompressInfoHeader);

	return retval;
 }




/*=========================================================================*/
/*  Decompress for MultiCPU(v2)                                            */
/*=========================================================================*/
unsigned int __stdcall CodecInst::Decompress3_thd(void* args)
{
	DBGPRINTOUT("---Start Decompress3Thread---\n")
	DecompressUnit *thd = (DecompressUnit *)args;
	thd->retval = (thd->pCodec)->Decompress3_Core2(thd->icinfo, thd->dwSize, thd->offset, thd->size, thd->height_st, thd->height);
	DBGPRINTOUT("---End Decompress3Thread---\n")
	_endthreadex(0);
	return 0;
}

DWORD CodecInst::Decompress3_Core1(ICDECOMPRESS* icinfo, DWORD dwSize) {
  // If you insert a Huffyuv-compressed AVI to a Premiere project and then
  // drag it on to the timeline, the following dialogue occurs:
  //
  // 1. Premiere calls ICDecompressBegin, asking Huffyuv to decompress
  //    to a bitmap with different dimensions than the compressed frame.
  //
  // 2. Huffyuv can't resize, so it returns ICERR_BADFORMAT.
  //
  // 3. Premiere calls ICDecompress without making another call to
  //    ICDecompressBegin.
  //
  // Therefore I now check for this case and compensate for Premiere's
  // negligence by making the DecompressBegin call myself.

  if (!decompressing) {
    DWORD retval = DecompressBegin(icinfo->lpbiInput, icinfo->lpbiOutput);
    if (retval != ICERR_OK)
      return retval;
  }

  // we change a top-down output request to a bottom-up one here. - bastel
  // (imho it should be possible to always do this, this code should never be reached with a negative height)
  if ( ignoretopdown ) icinfo->lpbiOutput->biHeight = abs(icinfo->lpbiOutput->biHeight);

  icinfo->lpbiOutput->biSizeImage = (icinfo->lpbiOutput->biWidth * icinfo->lpbiOutput->biHeight * icinfo->lpbiOutput->biBitCount) >> 3;

  return ICERR_OK;
}

DWORD CodecInst::Decompress3_Core2(ICDECOMPRESS* icinfo, DWORD dwSize, DWORD in_st, DWORD in_size, DWORD out_height_st, DWORD out_height) {
  //------------------------------------------------------------------------------
  LONG  out_biHeight = out_height;
  DWORD in_offs   = in_st;
  DWORD out_offs  = icinfo->lpbiOutput->biWidth * out_height_st * GetByteCount(icinfo->lpbiOutput);
  DWORD out_offs2 = icinfo->lpbiInput->biWidth * out_height_st * GetByteCount(icinfo->lpbiInput);
  unsigned char* u_decompress_yuy2_buffer = decompress_yuy2_buffer + out_offs2;
  //------------------------------------------------------------------------------

  int intype = GetBitmapType(icinfo->lpbiInput);
  if (intype < 0) {
    int method = GetMethod(icinfo->lpbiInput);
    int outtype = GetBitmapType(icinfo->lpbiOutput);

    if (decode_table_owner != this) {
      InitializeDecodeTables(GetEmbeddedHuffmanTable(icinfo->lpbiInput));
      decode_table_owner = this;
    }

	const unsigned long* const in = (unsigned long*)icinfo->lpInput + (in_offs/sizeof(unsigned long));
    unsigned char* const out = (unsigned char*)icinfo->lpOutput + out_offs;

	//Msg("intype=[%d], outtype=[%d]\n", intype, outtype);
    if (intype == -1) {   // decompressing HFYU16
      int stride = icinfo->lpbiOutput->biWidth * 2;
      const int size = stride * out_biHeight;
      // if image is interlaced, double stride so fields are treated separately
      if ( isinterlaced ) stride *= 2;

      unsigned char* const yuy2 = outtype>1 ? u_decompress_yuy2_buffer : out;

      if (method == methodMedian) {
        asm_DecompressHFYU16(in, yuy2, yuy2+size);
        asm_MedianRestore(yuy2, yuy2+size, stride);
      }
      else {
        asm_DecompressHFYU16Delta(in, yuy2, yuy2+size);
        if (method == methodGrad)
          mmx_RowAccum(yuy2, yuy2+size, stride);
      }

      if (outtype>1)    // HFYU16->RGB
        ConvertYUY2toRGB2(icinfo->lpbiOutput, yuy2, out, out_height);
    }
    else {   // decompressing HFYU24/HFYU32
      int stride = (icinfo->lpbiOutput->biWidth * icinfo->lpbiOutput->biBitCount) >> 3;
      const int size = stride * out_biHeight;
      // if image is interlaced, double stride so fields are treated separately
	  if ( isinterlaced ) stride *= 2;
      if (intype == -2) {   // HFYU24->RGB
        if (outtype == 4) {
          if (method == methodLeft || method == methodOld)
            asm_DecompressHFYU24To32Delta(in, out, out+size);
          else
            asm_DecompressHFYU24To32DeltaDecorrelate(in, out, out+size);
        } else {
          if (method == methodLeft || method == methodOld)
            asm_DecompressHFYU24To24Delta(in, out, out+size);
          else
            asm_DecompressHFYU24To24DeltaDecorrelate(in, out, out+size);
        }
      }
      else if (intype == -3) {    // HFYU32->RGBA
        if (method == methodLeft || method == methodOld)
          asm_DecompressHFYU32To32Delta(in, out, out+size);
        else
          asm_DecompressHFYU32To32DeltaDecorrelate(in, out, out+size);
      }
      else
        return ICERR_BADFORMAT;

      if ((method&~flagDecorrelate) == methodGrad)
        mmx_RowAccum(out, out+size, stride);
    }

    if (swapfields && isinterlaced )
      asm_SwapFields(out, out+((icinfo->lpbiOutput->biWidth * icinfo->lpbiOutput->biBitCount * out_biHeight) >> 3),
       (icinfo->lpbiOutput->biWidth * icinfo->lpbiOutput->biBitCount) >> 3);

    return ICERR_OK;
  }
  else if (intype == 1) {   // YUY2->RGB
    return ConvertYUY2toRGB2(icinfo->lpbiOutput, (unsigned char*)icinfo->lpInput+in_offs, (unsigned char*)icinfo->lpOutput+out_offs, out_height);
  }
  else if (intype == 2) {   // UYVY->RGB
    return ConvertUYVYtoRGB2(icinfo->lpbiOutput, (unsigned char*)icinfo->lpInput+in_offs, (unsigned char*)icinfo->lpOutput+out_offs, out_height);
  }
  else
    return ICERR_BADFORMAT;
}


DWORD CodecInst::Decompress3(ICDECOMPRESS* icinfo, DWORD dwSize) {
	DWORD retval = ICERR_OK;

	CompressInfoHeader* header = (CompressInfoHeader*)((unsigned char*)(icinfo->lpInput) + icinfo->lpbiInput->biSizeImage - sizeof(CompressInfoHeader));
	CompressInfo* info = (CompressInfo*)((unsigned char*)(icinfo->lpInput) + header->offset);;
	int cut_num = header->unitnum;
	if(cut_num > MAX_PROCESSOR_NUM) return ICERR_BADFORMAT;
	//Msg("---Decompress3[num=%d]---\n", cut_num);

	//---Ready decompress
	retval = Decompress3_Core1(icinfo, dwSize);
	if(retval != ICERR_OK)
		return retval;

	//---Start thread
	int i,m;
	int numThreads = 0;
	HANDLE hThread[MAX_PROCESSOR_NUM];
	DWORD height_num = header->unitsize;
	DecompressUnit decunit[MAX_PROCESSOR_NUM];
	DWORD height_offs = 0;
	bool rgb = false;
	if( (GetBitmapType(icinfo->lpbiInput)==1) || (GetBitmapType(icinfo->lpbiInput)==2) || (GetBitmapType(icinfo->lpbiInput)==-1 && GetBitmapType(icinfo->lpbiOutput)>1) )
		rgb = true;
	//bool rgb = (GetBitmapType(icinfo->lpbiOutput)>1) ? true : false;
	for(i=0,m=0; i<cut_num; i++){
		decunit[i].pCodec    = this;
		decunit[i].retval    = ICERR_OK;
		decunit[i].icinfo    = icinfo;
		decunit[i].dwSize    = dwSize;

		m = (rgb ? (cut_num-1-i) : i);
		decunit[i].offset    = info[m].code_start;
		decunit[i].size      = info[m].data_size;
		decunit[i].height_st = height_offs;
		decunit[i].height    = height_num;
		if(m==cut_num-1){
			decunit[i].height = icinfo->lpbiOutput->biHeight - (height_num * m);
		}
		height_offs += decunit[i].height;

		//---Start thread
		hThread[numThreads] = (HANDLE)_beginthreadex(NULL, 0, Decompress3_thd, &decunit[i], 0, NULL);
		SetThreadAffinityMask(hThread[numThreads], processor_mask[numThreads]);
		numThreads++;
		//retval = Decompress3_Core2(icinfo, dwSize, decunit[i].offset, decunit[i].size, decunit[i].height_st, decunit[i].height);
		//if(retval != ICERR_OK)
		//	return retval;
	}
	//---Wait multiple thread finish
	WaitForMultipleObjects(numThreads, hThread, TRUE, INFINITE);
	for(i=0; i<numThreads; i++){
		CloseHandle(hThread[i]);
		if(decunit[i].retval != ICERR_OK)  
			retval = decunit[i].retval;
	}
	return retval;
}
